/***************************************************************************
 ** This file is part of the lirc-0.5.4 package ****************************
 ** LIRC - Linux Infrared Remote Control ***********************************
 ***************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


/****************************************************************************
 ** lirc_parallel.c *********************************************************
 ****************************************************************************
 * 
 * lirc_parallel - device driver for infra-red signal receiving and
 *                 transmitting unit built by the author
 * 
 * Copyright (C) 1998 Christoph Bartelmus <columbus@hit.handshake.de>
 *
 */ 

/***********************************************************************
 *************************       Includes        ***********************
 ***********************************************************************/

#include <linux/version.h>
#if LINUX_VERSION_CODE >= 0x020100
#error "--- This Driver is not tested with 2.1.x and 2.2.x Kernels! ---"
#else

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/config.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/time.h>
#include <linux/mm.h>
#include <linux/delay.h>

#include <asm/io.h>
#include <asm/signal.h>
#include <asm/irq.h>

#include "hardware.h"
#include "lirc_parallel.h"

#define LIRC_DRIVER_NAME "lirc_parallel"

/***********************************************************************
 *************************   Globale Variablen   ***********************
 ***********************************************************************/

unsigned int irq = LIRC_IRQ;
unsigned int port = LIRC_PORT;
#ifdef LIRC_TIMER
unsigned int timer = 0;
unsigned int default_timer = LIRC_TIMER;
#endif

#define WBUF_SIZE (256)
#define RBUF_SIZE (256) /* this must be a power of 2 larger than 4 */

static unsigned long wbuf[WBUF_SIZE/4];
static unsigned long rbuf[RBUF_SIZE/4];

static struct wait_queue *lirc_wait=NULL;

unsigned int rptr=0,wptr=0;
unsigned int lost_irqs=0;

/***********************************************************************
 *************************   Interne Funktionen  ***********************
 ***********************************************************************/

unsigned int __inline__ in(int offset)
{
	return inb(port+offset);
}

void __inline__ out(int offset, int value)
{
	outb(value, port+offset);
}

unsigned int __inline__  lirc_get_timer()
{
	return(in(LIRC_PORT_TIMER)&LIRC_PORT_TIMER_BIT);
}

unsigned int __inline__  lirc_get_signal()
{
	return(in(LIRC_PORT_SIGNAL)&LIRC_PORT_SIGNAL_BIT);
}

void __inline__ lirc_on()
{
	out(LIRC_PORT_DATA,LIRC_PORT_DATA_BIT);
}

void __inline__ lirc_off()
{
	out(LIRC_PORT_DATA,0);
}

unsigned int init_lirc_timer()
{
	struct timeval tv,now;
	unsigned int level,newlevel,timeelapsed,newtimer;
	int count=0;
	
	do_gettimeofday(&tv);
	tv.tv_sec++;                     /* wait max. 1 sec. */
	level=lirc_get_timer();
	do
	{
		newlevel=lirc_get_timer();
		if(level==0 && newlevel!=0) count++;
		level=newlevel;
		do_gettimeofday(&now);
	}
	while(count<1000 && (now.tv_sec<tv.tv_sec 
			     || (now.tv_sec==tv.tv_sec 
				 && now.tv_usec<tv.tv_sec)));

	timeelapsed=((now.tv_sec+1-tv.tv_sec)*1000000+(now.tv_usec-tv.tv_usec));
	if(count>=1000 && timeelapsed>0)
	{
		if(default_timer==0)                    /* autodetect timer */
		{
			newtimer=(1000000*count)/timeelapsed;
			printk(KERN_INFO "%s: %u Hz timer detected\n",LIRC_DRIVER_NAME,newtimer);
			return(newtimer);
		}
		else
		{
			newtimer=(1000000*count)/timeelapsed;
			if(abs(newtimer-default_timer)>
			   default_timer/10) /* bad timer */
			{
				printk(KERN_NOTICE "%s: bad timer: %u Hz\n",
				       LIRC_DRIVER_NAME,newtimer);
				printk(KERN_NOTICE "%s: using default timer: %u Hz\n",
				       LIRC_DRIVER_NAME,default_timer);
				return(default_timer);
			}
			else
			{
				printk(KERN_INFO "%s: %u Hz timer detected\n",LIRC_DRIVER_NAME,newtimer);
				return(newtimer); /* use detected value */
			}
		}
	}
	else
	{
		printk(KERN_NOTICE "%s: no timer detected\n",LIRC_DRIVER_NAME);
		return(0);
	}
}

/***********************************************************************
 *************************   interrupt handler  ************************
 ***********************************************************************/

static inline void rbuf_write(unsigned long signal)
{
	unsigned int nwptr;

	nwptr=(wptr+4) & (RBUF_SIZE-1);
	if(nwptr==rptr) /* no new signals will be accepted */
	{
		lost_irqs++;
		printk(KERN_NOTICE "%s: buffer overrun\n",LIRC_DRIVER_NAME);
		return;
	}	
	*((unsigned long *) (((char *) rbuf )+wptr))=signal;
	wptr=nwptr;
}

void irq_handler(int i,void *blah,struct pt_regs * regs)
{
	struct timeval tv;
	static struct timeval lasttv;
	static int init=0;
	unsigned long signal;
	unsigned int level,newlevel;
	int timeout;

	/* disable interrupt */
	/*
	  disable_irq(irq);
	  out(LIRC_PORT_IRQ,in(LIRC_PORT_IRQ)&(~LP_PINTEN));
	*/
	if(in(1)&LP_PSELECD)
		return;

#ifdef LIRC_TIMER
	if(init)
	{
		do_gettimeofday(&tv);
		
	        signal=tv.tv_sec-lasttv.tv_sec;
		if(signal>15)
		{
			signal=0xFFFFFF;  /* really long time */
		}
		else
		{
			signal=signal*1000000+tv.tv_usec-lasttv.tv_usec
			+LIRC_SFH506_DELAY;
		};

		rbuf_write(signal); /* space */
	}
	else
	{
		if(timer==0) /* wake up; we'll lose this signal 
				but it will be garbage if the device 
				is turned on anyway
			      */
		{
			timer=init_lirc_timer();
			/* enable_irq(irq); */
			return;
		}
		init=1;
	}

	timeout=timer/10;           /* timeout after 1/10 sec. */
	signal=1;
	level=lirc_get_timer();
	do{
		newlevel=lirc_get_timer();
		if(level==0 && newlevel!=0) signal++;
		level=newlevel;

		/* giving up */
		if(signal>timeout || (in(1)&LP_PSELECD))
		{
			signal=0;
			printk(KERN_NOTICE "%s: timeout\n",LIRC_DRIVER_NAME);
			break;
		}
	}
	while(lirc_get_signal());
	if(signal!=0)
	{
		/* ajust value to usecs */
		signal=(unsigned int) ((unsigned long long) (signal*1000000)/timer);

		if(signal>LIRC_SFH506_DELAY)
		{
			signal-=LIRC_SFH506_DELAY;
		}
		else
		{
			signal=1;
		}
		rbuf_write(0x01000000|signal); /* pulse */
	}
	do_gettimeofday(&lasttv);
#else
		/* add your code here */
#endif
	
	wake_up_interruptible(&lirc_wait);

	/* enable interrupt */
	/*
	  enable_irq(irq);
	  out(LIRC_PORT_IRQ,in(LIRC_PORT_IRQ)|LP_PINTEN);
	*/
}

/***********************************************************************
 **************************   file_operations   ************************
 ***********************************************************************/

static int lirc_lseek(struct inode *inode,struct file *filep,off_t offset,int orig)
{
	return(-ESPIPE);
}

static int lirc_read(struct inode *node,struct file *filep,char *buf,int n)
{
	unsigned int minor=MINOR(node->i_rdev);
	int result;
	int count=0;
	
	switch(minor)
	{
	case LIRC_MINOR_BOTH:
	case LIRC_MINOR_RECEIVE:

		result=verify_area(VERIFY_WRITE,buf,n);
		if(result) return(result);

		while(count<n)
		{
			if(rptr!=wptr)
			{
				memcpy_tofs(buf+count,(char *) rbuf+rptr,1);
				rptr=(rptr+1)&(RBUF_SIZE-1);
				count++;
			}
			else
			{
				if(filep->f_flags & O_NONBLOCK)
				{
					result=-EAGAIN;
					break;
				}
				if(current->signal & ~current->blocked)
				{
					result=-EINTR;
					break;
				}
				interruptible_sleep_on(&lirc_wait);
				current->state = TASK_RUNNING;
			}
		}
		return(count ? count:result);
	case LIRC_MINOR_SEND:
		return(-EBADF);
	}
	return(-EBADF);
}

static int lirc_write(struct inode *node,struct file *filep,const char *buf,int n)
{
	unsigned int minor=MINOR(node->i_rdev);
	int result,count;
	unsigned int i;
	unsigned int level,newlevel,counttimer;
	unsigned long flags;
      
	switch(minor)
	{
	case LIRC_MINOR_BOTH:
	case LIRC_MINOR_SEND:
		
		if(n%sizeof(unsigned long)) return(-EINVAL);
		result=verify_area(VERIFY_READ,buf,n);
		if(result) return(result);

		count=n/sizeof(unsigned long);

		if(count>WBUF_SIZE || count%2==0) return(-EINVAL);

		memcpy_fromfs(wbuf,buf,n);

		/*
		  if(check_region(port,LIRC_PORT_LEN)!=0) return(-EBUSY);
		*/

#ifdef LIRC_TIMER
		if(timer==0) /* try again if device is ready */
		{
			timer=init_lirc_timer();
			if(timer==0) return(-EIO);
		}

		/* ajust values from usecs */
		for(i=0;i<count;i++) wbuf[i]=(unsigned long) ((unsigned long long) (wbuf[i]*timer)/1000000);
			
		save_flags(flags);cli();
		i=0;
		while(i<count)
		{
			level=lirc_get_timer();
			counttimer=0;
			lirc_on();
			do
			{
				newlevel=lirc_get_timer();
				if(level==0 && newlevel!=0) counttimer++;
				level=newlevel;
				if(in(1)&LP_PSELECD)
				{
					lirc_off();
					restore_flags(flags); /* sti(); */
					return(-EIO);
				}
			}
			while(counttimer<wbuf[i]);i++;

			lirc_off();
			if(i==count) break;

			counttimer=0;
			do
			{
				newlevel=lirc_get_timer();
				if(level==0 && newlevel!=0) counttimer++;
				level=newlevel;
				if(in(1)&LP_PSELECD)
				{
					restore_flags(flags); /* sti(); */
					return(-EIO);
				}
			}
			while(counttimer<wbuf[i]);i++;
		}
		restore_flags(flags); /* sti(); */
#else
		/* 
		   place code that handles write
		   without extarnal timer here
		*/
#endif
		return(n);
	case LIRC_MINOR_RECEIVE:
		return(-EBADF);
	}
	return(-EBADF);
}

static int lirc_select(struct inode *node,struct file *file,int sel_type,select_table *wait)
{
	unsigned int minor=MINOR(node->i_rdev);
  
	switch(minor)
	{
	case LIRC_MINOR_BOTH:
	case LIRC_MINOR_RECEIVE:
		if(sel_type!=SEL_IN)
			return 0;
		if(rptr!=wptr)
			return(1);
		
		select_wait(&lirc_wait,wait);
		return(0);
	case LIRC_MINOR_SEND:
		return(0);
	}
	return(0);
}

static int lirc_ioctl(struct inode *node,struct file *filep,unsigned int cmd,unsigned long arg)
{
	unsigned int minor=MINOR(node->i_rdev);
	int result;
	int lirc_state;
	int timeout,level,newlevel;
	unsigned int dhl,dlh;
	
	switch(minor)
	{
	case LIRC_MINOR_SEND:
		switch(cmd)
		{
		case LIRC_GET_DHL:
			result=verify_area(VERIFY_WRITE,(unsigned int *) arg,sizeof(unsigned int));
			if(result) return(result);

			if(timer==0) /* try again if device is ready */
			{
				timer=init_lirc_timer();
				if(timer==0) return(-EIO);
			}

			lirc_on();
			timeout=timer/10;
			level=lirc_get_timer();

			while(!lirc_get_signal())
			{
				newlevel=lirc_get_timer();
				if(level==0 && newlevel!=0) timeout--;
				level=newlevel;
				
				if(timeout<0 || (in(1)&LP_PSELECD))
				{
					lirc_off();
					return(-EIO);
				}
			}
			lirc_off();
			timeout=timer/10;
			level=lirc_get_timer();
			
			while(lirc_get_signal())
			{
				newlevel=lirc_get_timer();
				if(level==0 && newlevel!=0) timeout--;
				level=newlevel;
				
				if(timeout<0 || (in(1)&LP_PSELECD))
				{
					return(-EIO);
				}
			}
			dhl=(timer/10)-timeout;
			dhl=(int) ((unsigned long long) (dhl*1000000)/timer);
			put_user(dhl,(unsigned int *) arg);
			break;
		case LIRC_GET_DLH:
			result=verify_area(VERIFY_WRITE,(unsigned int *) arg,sizeof(unsigned int));
			if(result) return(result);

			if(timer==0) /* try again if device is ready */
			{
				timer=init_lirc_timer();
				if(timer==0) return(-EIO);
			}

			lirc_off();
			timeout=timer/10;
			level=lirc_get_timer();

			while(lirc_get_signal())
			{
				newlevel=lirc_get_timer();
				if(level==0 && newlevel!=0) timeout--;
				level=newlevel;
				
				if(timeout<0 || (in(1)&LP_PSELECD))
				{
					return(-EIO);
				}
			}
			lirc_on();
			timeout=timer/10;
			level=lirc_get_timer();
			
			while(!lirc_get_signal())
			{
				newlevel=lirc_get_timer();
				if(level==0 && newlevel!=0) timeout--;
				level=newlevel;
				
				if(timeout<0 || (in(1)&LP_PSELECD))
				{
					lirc_off();
					return(-EIO);
				}
			}
			lirc_off();
			dlh=(timer/10)-timeout;
			dlh=(int) ((unsigned long long) (dlh*1000000)/timer);
			put_user(dlh,(unsigned int *) arg);
			break;
		default:
			return(-ENOIOCTLCMD);
		}
		return(0);
	case LIRC_MINOR_BOTH:
	case LIRC_MINOR_RECEIVE:
		switch (cmd)
		{
		case LIRC_GETLOSTIRQS:
			result=verify_area(VERIFY_WRITE,(unsigned int *) arg,sizeof(unsigned int));
			if(result) return(result);
			put_user(lost_irqs,(unsigned int *) arg);
			break;
		case LIRC_DISABLEPRINTER:
			outb(2,LP_PSELECP);
			break;
		case LIRC_GETSTATE:
			result=verify_area(VERIFY_WRITE,(int *) arg,sizeof(int));
			if(result) return(result);
			if(in(1)&LP_PSELECD) lirc_state=LIRC_OFFLINE;
			else lirc_state=LIRC_ONLINE;
			put_user(lirc_state,(int *) arg);
			break;		
		default:
			return(-ENOIOCTLCMD);
		}
		return(0);		
	}
	return(-EBADF);
}

static int lirc_open(struct inode* node,struct file* filep)
{
	unsigned int minor = MINOR(node->i_rdev);
	int result;
	
	switch(minor)
	{
	case LIRC_MINOR_BOTH:                /* uses interrupts exclusively */
	case LIRC_MINOR_RECEIVE:
		result=request_irq(irq,irq_handler,SA_INTERRUPT,LIRC_DRIVER_NAME,NULL);
		switch(result)
		{
		case -EBUSY:
			printk(KERN_NOTICE "%s: IRQ %u busy\n",LIRC_DRIVER_NAME,irq);
			return(-EBUSY);
		case -EINVAL:
			printk(KERN_NOTICE "%s: bad irq number or handler\n",LIRC_DRIVER_NAME);
			return(-EINVAL);
		}
		/* enable interrupt */
		out(2,LP_PINITP|LP_PSELECP|LP_PINTEN);

		/* init read ptr */
		rptr=wptr=0;
		lost_irqs=0;
		break;
	case LIRC_MINOR_SEND:
		break;
	default:
		return(-ENODEV);
	}
	MOD_INC_USE_COUNT;
	return(0);
}

static void lirc_close(struct inode* node,struct file* filep)
{
	unsigned int minor = MINOR(node->i_rdev);
	
	switch(minor)
	{
	case LIRC_MINOR_BOTH:                /* uses interrupts exclusively */
	case LIRC_MINOR_RECEIVE:
		free_irq(irq,NULL);

		/* disable interrupt */
		/*
		  FIX ME
		out(2,LP_PINITP|LP_PSELECP);
		*/
		break;
	case LIRC_MINOR_SEND:
		break;
	}
	MOD_DEC_USE_COUNT;
}

static struct file_operations lirc_fops = 
{
	lirc_lseek,	 /* lseek */
	lirc_read,   	 /* read */
	lirc_write,       /* write */
	NULL,		 /* readdir */
	lirc_select,      /* select */
	lirc_ioctl,
	NULL,            /* mmap  */
	lirc_open,
	lirc_close,
	NULL,            /* fsync */
	NULL,            /* fasync */
	NULL,            /* check_media_change */
	NULL             /* revalidate */
};

/***********************************************************************
 ******************   init_module()/cleanup_module()  ******************
 ***********************************************************************/

int init_module(void)
{
	if(check_region(port,LIRC_PORT_LEN)!=0)
	{
		printk(KERN_NOTICE "%s: port already in use\n",LIRC_DRIVER_NAME);
		return(-EBUSY);
	}
	request_region(port,LIRC_PORT_LEN,LIRC_DRIVER_NAME);

#ifdef LIRC_TIMER
#       ifdef DEBUG
	out(LIRC_PORT_DATA,LIRC_PORT_DATA_BIT);
#       endif
	
	timer=init_lirc_timer();

#       if 0 	/* continue even if device is offline */
	if(timer==0) 
	{
		release_region(port,LIRC_PORT_LEN);
		return(-EIO);
	}
	
#       endif
#       ifdef DEBUG
	out(LIRC_PORT_DATA,0);
#       endif
#endif	
	if(register_chrdev(LIRC_MAJOR, LIRC_DRIVER_NAME, &lirc_fops)<0)
	{
		printk(KERN_NOTICE "%s: register_chrdev() failed\n",LIRC_DRIVER_NAME);
		release_region(port,LIRC_PORT_LEN);
		return(-EIO);
	}
	printk(KERN_INFO "%s: installed using port 0x%04x irq %d\n",LIRC_DRIVER_NAME,port,irq);
	return(0);
}
  
void cleanup_module(void)
{
	if(MOD_IN_USE) return;
	release_region(port,LIRC_PORT_LEN);
	unregister_chrdev(LIRC_MAJOR,LIRC_DRIVER_NAME);
}
#endif
